import dotenv
from langchain import chat_models
from langchain_community.adapters import openai as openai_compatible
from pydantic import Field, field_validator

from contextcheck.connectors.connector import ConnectorBase

dotenv.load_dotenv()


class ConnectorOpenAICompatible(ConnectorBase):
    model: str = Field(default="gpt-4o-mini", description="A model used for a request")
    provider: str = Field(default="ChatOpenAI", description="A name of the provider")
    temperature: float | None = Field(default=None, description="Temperature used for the llm")
    max_tokens: int | None = Field(default=None, description="Max tokens for llm output")

    @field_validator("provider", mode="before")
    @classmethod
    def validated_provider(cls, provider: str) -> str:
        if provider is not None:
            if provider not in chat_models.__all__:
                raise ValueError(
                    f"Provider '{provider}' not found. Choose one of: {chat_models.__all__}"
                )

        return provider

    def send(self, data: dict) -> dict:
        kwargs = {
            "messages": [data],
            "model": self.model,
            "provider": self.provider,
        }

        if self.temperature:
            kwargs["temperature"] = self.temperature

        if self.max_tokens:
            kwargs["max_tokens"] = self.max_tokens

        chat_completion = openai_compatible.chat.completions.create(**kwargs)
        return dict(chat_completion)



from importlib import import_module


def import_class_from_string(path: str):
    """
    Import class from string path.

    Args:
        path (str): Path to class.

    Returns:
        class: Imported class.
    """
    try:
        module_path, class_name = path.rsplit(".", 1)
        module = import_module(module_path)
        clazz = getattr(module, class_name)
        return clazz
    except (ValueError, AttributeError, ModuleNotFoundError):
        raise ImportError(f"Could not import class from path: {path}")



from pydantic import Field, model_serializer

from contextcheck.connectors.connector_http import ConnectorHTTP
from contextcheck.endpoints.endpoint import EndpointBase, EndpointConfig
from contextcheck.models.request import RequestBase


# TODO: Add Fields with defaults and descriptions
class EndpointCCConfig(EndpointConfig):
    top_k: int = 3
    use_ranker: bool = True
    collection_name: str = "default"
    url: str = ""  # AnyUrl type can be applied
    additional_headers: dict = {}


class EndpointCC(EndpointBase):
    config: EndpointCCConfig = Field(
        default_factory=EndpointCCConfig, description="Configuration for CC ednpoint"
    )

    class RequestModel(RequestBase):
        @model_serializer
        def serialize(self) -> dict:
            # include possible request fields prompt for QA, query for semantic search
            return {"prompt": self.message, "query": self.message}

    def model_post_init(self, __context) -> None:
        self.connector = ConnectorHTTP(
            url=self.config.url,
            additional_headers=self.config.additional_headers,
        )



from typing import Self

from pydantic import BaseModel, ConfigDict, model_validator
from pydantic_core import from_json

fields_computation_map = {"eval": lambda x, context: eval(str(x), context)}


def replace_str_with_json(d: dict) -> dict:
    """Replace all strings JSON-parsable with corresponding python object."""

    if "parse_response_as_json" in d:
        for k, v in d.items():
            if isinstance(v, str) and check_beginning_bracket(v):  # Clumsy way to c
                try:
                    d[k] = from_json(v)
                except ValueError:  # Means that string is not json, leave it alone
                    pass
            elif isinstance(v, dict):
                d[k] = replace_str_with_json(v)
            else:
                continue

    return d


def check_beginning_bracket(s: str) -> bool:
    """Check if string starts with bracket."""
    try:
        return s.strip()[0] in ("{", "[")
    except IndexError:
        return False


class RequestBase(BaseModel):
    model_config = ConfigDict(extra="allow")
    message: str | dict | None = None

    @model_validator(mode="before")
    @classmethod
    def from_obj(cls, obj: str | dict) -> dict:
        """
        If request is a string, return it as message field.
        If request is a dict, dig deeper for string and convert it to dict if json.
        """
        return replace_str_with_json(obj) if isinstance(obj, dict) else {"message": obj}

    def build(self, context: dict | None = None) -> Self:
        """Build request based on prototype values."""

        def _search_and_replace(d: dict) -> dict:
            """Search recursively dict for computation string and replace it with function result."""
            # NOTE: Hmm, What if d = {"eval": "1 == 2", "value1": {"eval": "1 == 2"}}
            # then the first eval would make d of type bool and d[key] = ... would raise an error
            # Unless the result of an eval is not a dict then this will cause problems
            # TODO: Fix it if time allows or forbid too deep nesting i.e. one level should be ok
            for key, value in d.items():
                if key in fields_computation_map:
                    d = fields_computation_map[key](value, context)
                elif isinstance(value, dict):
                    d[key] = _search_and_replace(value)
            return d

        return self.model_validate(_search_and_replace(self.model_dump()))



import os
import sys
from datetime import datetime, timezone
from typing import Literal

from loguru import logger
from pydantic import BaseModel, field_validator, model_validator

from contextcheck.executors.executor import Executor
from contextcheck.interfaces.interface import InterfaceBase
from contextcheck.interfaces.interface_output_file import InterfaceOutputFile
from contextcheck.interfaces.interface_tui import InterfaceTUI
from contextcheck.models.models import TestScenario


class TestsRouter(BaseModel):
    output_type: Literal["console", "file"]
    filename: list[str] | None = []
    folder: str | None = None
    output_folder: str | None = None
    exit_on_failure: bool = False
    global_test_timestamp: str | None = None
    aggregate_results: bool = False
    show_time_statistics: bool = False

    @field_validator("filename")
    @classmethod
    def check_filename(cls, value: list[str]) -> list[str]:
        if value:
            invalid_files = [file for file in value if not os.path.isfile(file)]
            if invalid_files:
                raise ValueError(
                    f'Files {", ".join(invalid_files)} do not exist. Check --filename argument.'
                )
        return value

    @field_validator("folder")
    @classmethod
    def check_folder(cls, value: str) -> str:
        if value and not os.path.isdir(value):
            raise ValueError(f'Folder "{value}" does not exist. Check --folder argument.')
        return value

    @model_validator(mode="before")
    @classmethod
    def check_files_and_folders(cls, data: dict) -> dict:
        if data["output_type"] == "file":
            if not os.path.isdir(data["output_folder"]):
                os.makedirs(data["output_folder"])
        return data

    def model_post_init(self, __context) -> None:
        if not self.global_test_timestamp:
            now = datetime.now(timezone.utc)
            self.global_test_timestamp = str(datetime.timestamp(now))

    def _run_test_scenario(self, filename: str, interface_type: InterfaceBase) -> bool | None:
        ui = interface_type(test_scenario_filename=filename)  # type: ignore
        ts = TestScenario.from_yaml(ui.get_scenario_path())

        executor = Executor(ts, ui=ui, exit_on_failure=self.exit_on_failure)
        scenario_result = executor.run_all()
        executor.summary(
            output_folder=self.output_folder,
            global_test_timestamp=self.global_test_timestamp,
            aggregate_results=self.aggregate_results,
            show_time_statistics=self.show_time_statistics,
        )

        return scenario_result, executor.early_stop

    def run_tests(self):
        scenario_results = []
        type_map = {"console": InterfaceTUI, "file": InterfaceOutputFile}
        interface_type = type_map[self.output_type]

        filenames = []
        if self.filename:
            for filename in self.filename:
                filenames.append(filename)
        elif self.folder:
            for filename in os.listdir(self.folder):
                if filename.endswith(".yaml"):
                    filenames.append(filename)

        if not filenames:
            logger.warning("No test scenario to run")

        # TODO: Potential place to increase performance by running several scenarios at once
        for filename in filenames:
            scenario_result, early_stop = self._run_test_scenario(filename, interface_type)
            scenario_results.append(scenario_result)
            if self.exit_on_failure and early_stop:
                break

        if self.exit_on_failure and not all(scenario_results):
            sys.exit(1)



import json

import jsonschema
import jsonschema.exceptions
from loguru import logger
from pydantic import BaseModel


class JsonValidator(BaseModel):
    request_json: str
    assertion_schema: dict | None = None

    def is_valid(self):
        try:
            json.loads(self.request_json)
            return True
        except json.JSONDecodeError as e:
            logger.error(f"Error message: {e}")
            return False

    def has_valid_schema(self):
        if not self.is_valid():
            return False

        if self.assertion_schema is None:
            raise ValueError("Assertion schema is not provided.")

        try:
            jsonschema.validate(json.loads(self.request_json), schema=self.assertion_schema)
            return True
        except (jsonschema.exceptions.ValidationError, jsonschema.exceptions.SchemaError) as e:
            logger.error(f"Error message: {e}")
            return False



import sys
from argparse import ArgumentParser
from pathlib import Path

from contextcheck.generators.generate_answers import AnswerGenerator
from contextcheck.generators.utils import import_class_from_string


def generate_answers(
    output_file: str,
    wrapper_class_path: str,
    top_k: int,
    questions_file: Path,
    use_ranker: bool,
    collection_name: str,
    debug: bool,
) -> None:
    """
    Generate answers for the given questions and save them to a YAML file.

    Args:
        @param output_file: Output path and filename to save the generated questions.
        @param wrapper_class_path: Full path to the API wrapper class definition.
        @param top_k: Number of answers.
        @param questions_file: Path to the questions file.

    Returns:
        None
    """
    api_wrapper = import_class_from_string(wrapper_class_path)

    generator_args = {
        "api_wrapper": api_wrapper(),
        "top_k": top_k,
        "questions_file": questions_file,
        "use_ranker": use_ranker,
        "collection_name": collection_name,
        "debug": debug,
    }

    generator = AnswerGenerator(**generator_args)
    generator.save_to_yaml(output_file)


def main():
    parser = ArgumentParser(
        prog="question_answering.py", description="Refer to readme for more information."
    )
    parser.add_argument(
        "--questions-file",
        type=str,
        required=True,
        help="Path to the YAML file containing the questions.",
    )
    parser.add_argument(
        "--output-file",
        type=str,
        required=True,
        help="Output path and filename to save the generated questions. "
        "Example: 'my_folder/my_file.yaml'. If folder does not exist, it will be created.",
    )
    parser.add_argument(
        "--wrapper-class-path",
        type=str,
        required=True,
        help="Full path to the API wrapper class definition. "
        "Example: For a class 'ClassName' defined in file 'module/submodule.py' use format"
        " 'module.submodule.ClassName'.",
    )
    parser.add_argument(
        "--top-k",
        type=int,
        default=3,
        help="Number of answers",
    )
    parser.add_argument(
        "--use-ranker",
        type=bool,
        default=True,
        help="Use ranker to get relevant documents. Default is True.",
    )
    parser.add_argument(
        "--collection-name",
        type=str,
        default="default",
        help="Name of the collection to use. Default is 'default'.",
    )

    parser.add_argument(
        "--debug",
        type=bool,
        action="store_true",
        help="It will append semantic search chunks along with the questions",
    )

    args = parser.parse_args(args=None if sys.argv[1:] else ["--help"])

    generate_answers(
        output_file=args.output_file,
        wrapper_class_path=args.wrapper_class_path,
        top_k=args.top_k,
        questions_file=Path(args.questions_file),
        use_ranker=args.use_ranker,
        collection_name=args.collection_name,
        debug=args.debug,
    )


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod

from pydantic import BaseModel, Field


# TODO: For list[dict] return types I'd create a pydantic model, as otherwise it could be hard
# to work with
class RagApiWrapperBase(BaseModel, ABC):
    timeout: int = Field(default=60, description="A timeout set for a request")

    @abstractmethod
    def list_documents(self) -> list[dict[str, str]]:
        """
        List documents from the RAG API.
        Returns a list of dictionaries, each containing the document ID and the document title.
        """

    @abstractmethod
    def get_document_chunks(self, document_id: str) -> list[str]:
        """
        Retrieve document chunks by document ID from the RAG API.
        Returns a list of strings, each representing a chunk of the document.
        Chunks should be ordered as they appear in the document.
        """

    @abstractmethod
    def query_semantic_db(self, query: str, **kwargs) -> list[dict]:
        """
        Query the semantic database through the RAG API.
        Returns a list of dicts, each representing a semantic search result.
        """

    @abstractmethod
    def query_qa(self, query: str, **kwargs) -> dict:
        """
        Query the question answering model through the RAG API.
        Returns a dict, representing an answer to the query.
        """



from contextcheck.endpoints.endpoint import EndpointBase
from contextcheck.endpoints.endpoint_cc import EndpointCC
from contextcheck.endpoints.endpoint_config import EndpointConfig, EndpointsEnum
from contextcheck.endpoints.endpoint_dummy_echo import EndpointDummyEcho
from contextcheck.endpoints.endpoint_openai_compatible import EndpointOpenAICompatible
from contextcheck.endpoints.endpoint_tg_chatbot import EndpointTGChatBot

ENDPOINT_MAPPING: dict[EndpointsEnum, EndpointBase] = {
    EndpointsEnum.OPENAI: EndpointOpenAICompatible,
    EndpointsEnum.OPENAI_COMPATIBLE: EndpointOpenAICompatible,
    EndpointsEnum.TG_CHATBOT: EndpointTGChatBot,
    EndpointsEnum.ECHO: EndpointDummyEcho,
    EndpointsEnum.CC_PROMPT_LLM: EndpointCC,
    EndpointsEnum.CC_SS: EndpointCC,
}


def factory(endpoint_config: EndpointConfig) -> EndpointBase:
    kind = endpoint_config.kind
    try:
        endpoint_class = ENDPOINT_MAPPING[kind]
    except KeyError:
        raise ValueError("No endpoint for this kind of config.")

    return endpoint_class(config=endpoint_config.model_dump())



import time

from pydantic import BaseModel


class ConnectorStats(BaseModel):
    conn_start_time: float | None = None
    conn_end_time: float | None = None
    conn_duration: float | None = None


class ConnectorBase(BaseModel):
    stats: ConnectorStats = ConnectorStats()

    def __enter__(self):
        self.stats.conn_start_time = time.perf_counter()
        self.stats.conn_end_time = None
        return self

    def __exit__(self, type, value, traceback):
        self.stats.conn_end_time = time.perf_counter()
        self.stats.conn_duration = self.stats.conn_end_time - self.stats.conn_start_time  # type: ignore

    def send(self, *args, **kwargs) -> dict:
        raise NotImplementedError



import re
from enum import StrEnum

from pydantic import BaseModel, ConfigDict, Field, model_validator

from contextcheck.assertions.llm_metrics import (
    LlmMetricEnum,
    LLMMetricEvaluator,
    llm_metric_factory,
)
from contextcheck.assertions.utils import JsonValidator
from contextcheck.endpoints.endpoint import EndpointBase
from contextcheck.models.request import RequestBase
from contextcheck.models.response import ResponseBase


class AssertionBase(BaseModel):
    model_config = ConfigDict(extra="allow")
    result: bool | None = Field(default=None, description="Result of a assertion")

    @model_validator(mode="before")
    @classmethod
    def from_obj(cls, obj: dict | str) -> dict:
        # Default assertion without keyword:
        return obj if isinstance(obj, dict) else {"eval": obj}

    def __call__(
        self, request: RequestBase, response: ResponseBase, eval_endpoint: EndpointBase
    ) -> bool:
        raise NotImplementedError


class AssertionEval(AssertionBase):
    eval: str = Field(description="Eval string to be used for python's eval")

    def __call__(
        self, request: RequestBase, response: ResponseBase, eval_endpoint: EndpointBase
    ) -> bool:
        if self.result is None:
            try:
                # NOTE: I suppose running eval has some major security risks attached
                result = eval(self.eval)
            except NameError:
                raise NameError(f"Given eval `{self.eval}` uses non-existent name.")
            if not isinstance(result, bool):
                raise ValueError(f"Given eval `{self.eval}` does not evaluate to bool.")
            self.result = result
        return self.result


class AssertionLLM(AssertionBase):
    llm_metric: LlmMetricEnum
    reference: str = Field(default="", description="Reference given to llm")
    assertion: str = Field(
        default="", description="Assertion given to llm to compare against the obtained results"
    )

    def __call__(
        self, request: RequestBase, response: ResponseBase, eval_endpoint: EndpointBase
    ) -> bool:
        if self.result is None:
            # This is required as otherwise time statistics will be shared across assertions
            # for a given TestScenario
            eval_endpoint = eval_endpoint.model_copy(deep=True)
            metric = llm_metric_factory(metric_type=self.llm_metric)

            self.metric_evaluator = LLMMetricEvaluator(
                eval_endpoint=eval_endpoint, metric=metric  # type: ignore
            )

            self.result = self.metric_evaluator.evaluate(
                input=request.message,  # type: ignore
                output=response.message,  # type: ignore
                **self.model_dump(exclude={"llm_metric", "result"}),
            )

        return self.result


class DeterministicMetricsEnum(StrEnum):
    CONTAINS = "contains"
    ICONTAINS = "icontains"
    CONTAINS_ALL = "contains-all"
    ICONTAINS_ALL = "icontains-all"
    CONTAINS_ANY = "contains-any"
    ICONTAINS_ANY = "icontains-any"
    IS_VALID_JSON = "is-valid-json"
    HAS_VALID_JSON_SCHEMA = "has-valid-json-schema"
    EQUALS = "equals"
    REGEX = "regex"


DETERMINISTIC_METRICS_MAPPING = {
    DeterministicMetricsEnum.CONTAINS: lambda assertion, response: assertion in response,
    DeterministicMetricsEnum.ICONTAINS: lambda assertion, response: assertion.lower()
    in response.lower(),
    DeterministicMetricsEnum.CONTAINS_ALL: lambda assertion, response: all(
        [assertion in response for assertion in assertion]
    ),
    DeterministicMetricsEnum.ICONTAINS_ALL: lambda assertion, response: all(
        [assertion.lower() in response.lower() for assertion in assertion]
    ),
    DeterministicMetricsEnum.CONTAINS_ANY: lambda assertion, response: any(
        [assertion in response for assertion in assertion]
    ),
    DeterministicMetricsEnum.ICONTAINS_ANY: lambda assertion, response: any(
        [assertion.lower() in response.lower() for assertion in assertion]
    ),
    DeterministicMetricsEnum.IS_VALID_JSON: lambda assertion, response: JsonValidator(
        request_json=response
    ).is_valid(),
    DeterministicMetricsEnum.HAS_VALID_JSON_SCHEMA: lambda assertion, response: JsonValidator(
        request_json=response, assertion_schema=assertion
    ).has_valid_schema(),
    DeterministicMetricsEnum.EQUALS: lambda assertion, response: assertion == response,
    DeterministicMetricsEnum.REGEX: lambda assertion, response: bool(re.match(assertion, response)),
}


class AssertionDeterministic(AssertionBase):
    kind: DeterministicMetricsEnum = Field(description="A type of deterministic evaluation")
    assertion: str | list[str] | dict | None = Field(
        default=None, description="Assertion to be evaluated"
    )

    def __call__(
        self, request: RequestBase, response: ResponseBase, eval_endpoint=EndpointBase
    ) -> bool:
        if self.result is None:

            if self.kind in DETERMINISTIC_METRICS_MAPPING:
                self.result = DETERMINISTIC_METRICS_MAPPING[self.kind](
                    self.assertion, response.message
                )
            else:
                raise ValueError(f"Given kind `{self.kind}` is not a deterministic metric.")

        return self.result  # type: ignore
